# Natural Language Toolkit: Chart Parser Demo
#
# Copyright (C) 2001 University of Pennsylvania
# Author: Edward Loper <edloper@gradient.cis.upenn.edu>
# URL: <http://nltk.sf.net>
# For license information, see LICENSE.TXT
#
# $Id: chart.py,v 1.36 2003/08/09 19:39:04 edloper Exp $

"""
A graphical tool for exploring chart parsing.

Chart parsing is a flexible parsing algorithm that uses a data
structure called a "chart" to record hypotheses about syntactic
constituents.  Each hypothesis is represented by a single "edge" on
the chart.  A set of "chart rules" determine when new edges can be
added to the chart.  This set of rules controls the overall behavior
of the parser (e.g., whether it parses top-down or bottom-up).

The chart parsing tool demonstrates the process of parsing a single
sentence, with a given grammar and lexicon.  Its display is divided
into three sections: the bottom section displays the chart; the middle
section displays the sentence; and the top section displays the
partial syntax tree corresponding to the selected edge.  Buttons along
the bottom of the window are used to control the execution of the
algorithm.

The chart parsing tool allows for flexible control of the parsing
algorithm.  At each step of the algorithm, you can select which rule
or strategy you wish to apply.  This allows you to experiment with
mixing different strategies (e.g., top-down and bottom-up).  You can
exercise fine-grained control over the algorithm by selecting which
edge you wish to apply a rule to.
"""

# At some point, we should rewrite this tool to use the new canvas
# widget system.

import Tkinter
import math

from nltk.parser.chart import *
from nltk.cfg import *
from nltk.tokenizer import WSTokenizer
from nltk.token import Token, Location
from nltk.tree import TreeToken
from nltk.draw import ShowText, EntryDialog
from nltk.draw.cfg import CFGEditor

# Known bug: ChartView doesn't handle edges generated by epsilon
# productions (e.g., [Production: PP -> ]) very well.

class ProductionView:
    """
    View a list of productions.  This can be used for the lexicon or for the
    grammar.
    """
    def __init__(self, productions, title='Production List'):
        self._productions = productions

        self._top = Tkinter.Tk()
        self._top.title(title)

        # Add a button list.
        buttons = Tkinter.Frame(self._top)
        buttons.pack(side='bottom')
        Tkinter.Button(buttons, text='Done', command=self.destroy).pack(side='right')

        # Create the list.
        listframe = Tkinter.Frame(self._top)
        listframe.pack(fill='both', expand='y')
        listscroll = Tkinter.Scrollbar(listframe, orient='vertical')
        self._listbox = listbox = Tkinter.Listbox(listframe)
        listbox.config(yscrollcommand = listscroll.set)
        listscroll.config(command=listbox.yview)
        listscroll.pack(side='right', fill='y')
        listbox.pack(side='left', fill='both', expand=1)

        # Add the productions.  We'll assume the production list is immutable..
        for production in self._productions:
            listbox.insert('end', production)

        # Add some keyboard bindings.
        #listbox.bind("<Double-Button-1>", self._ok)
        listbox.bind("q", self.destroy)

    def destroy(self, *args):
        if self._top == None: return
        self._top.destroy()
        self._top = None

#//////////////////////////////////////////////////////////////////////
# Small Chart Matrix View
#//////////////////////////////////////////////////////////////////////

class SmallChartMatrixView:
    """
    A view of a chart that displays the contents of the corresponding matrix.
    """
    def __init__(self, parent, chart, source):
        self._chart = chart
        self._source = source

        self._root = Tkinter.Toplevel(parent)
        self._root.title('Chart Matrix')
#        def destroy1(e, top=self._root): top.destroy()
#        def destroy2(top=self._root): top.destroy()
#        self._root.bind('q', destroy1)
#        b = Tkinter.Button(self._root, text='Close', command=destroy2)
#        b.pack(side='bottom')

        self._init_locmap()
        self._canvas = Tkinter.Canvas(self._root, width=200, height=200,
                                      background='white')
        self._canvas.pack(expand=1, fill='both')
        self._init_locmap()
        self.draw()

    def _init_locmap(self):
        # Mapping from indices to locations
        self._index2loc = [t.loc().start() for t in self._source]
        self._index2loc.append(self._source[-1].loc().end())

        # Mapping from locations to indices
        self._loc2index = {}
        for i in self._index2loc:
            self._loc2index[self._index2loc[i]] = i

    def update(self, chart, source=None):
        self._chart = chart
        if source is not None:
            self._source = source
            self._init_locmap()
            self.draw()
            
        # Gray out everything:
        N = len(self._source)+1
        for i in range(N):
            for j in range(i, N):
                self._canvas.itemconfig(self._cells[i][j], fill='gray20')
        
        for edge in self._chart.edges():
            self.color_edge(edge, 'blue4')
            i = self._loc2index[edge.loc().start()]
            j = self._loc2index[edge.loc().end()]
            self._canvas.itemconfig(self._cells[i][j], fill='blue2')

    def color_edge(self, edge, color):
        i = self._loc2index[edge.loc().start()]
        j = self._loc2index[edge.loc().end()]
        self._canvas.itemconfig(self._cells[i][j], fill=color)

    def draw(self):
        LEFT_MARGIN = BOT_MARGIN = 15
        TOP_MARGIN = 5
        c = self._canvas
        c.delete('all')
        N = len(self._source)+1
        dx = (int(c['width'])-LEFT_MARGIN)/N
        dy = (int(c['height'])-TOP_MARGIN-BOT_MARGIN)/N

        c.delete('all')

        # Labels and dotted lines
        for i in range(N):
            c.create_text(LEFT_MARGIN-2, i*dy+dy/2+TOP_MARGIN,
                          text=`self._index2loc[i]`, anchor='e')
            c.create_text(i*dx+dx/2+LEFT_MARGIN, N*dy+TOP_MARGIN+1,
                          text=`self._index2loc[i]`, anchor='n')
            c.create_line(LEFT_MARGIN, dy*(i+1)+TOP_MARGIN, 
                          dx*N+LEFT_MARGIN, dy*(i+1)+TOP_MARGIN, dash='.')
            c.create_line(dx*i+LEFT_MARGIN, TOP_MARGIN,
                          dx*i+LEFT_MARGIN, dy*N+TOP_MARGIN, dash='.')

        # A box around the whole thing
        c.create_rectangle(LEFT_MARGIN, TOP_MARGIN,
                           LEFT_MARGIN+dx*N, dy*N+TOP_MARGIN,
                           width=2)

        self._cells = []
        for i in range(N):
            self._cells.append([None]*i) # Extra padding
            for j in range(i, N):
                t = c.create_rectangle(j*dx+LEFT_MARGIN, i*dy+TOP_MARGIN,
                                       (j+1)*dx+LEFT_MARGIN,
                                       (i+1)*dy+TOP_MARGIN,
                                       fill='gray20')
                self._cells[-1].append(t)

        self.update(self._chart)

#//////////////////////////////////////////////////////////////////////
# Chart View
#//////////////////////////////////////////////////////////////////////

class ChartView:
    """
    A component for viewing charts.  This is used by C{ChartDemo} to
    allow students to interactively experiment with various chart
    parsing techniques.  It is also used by C{Chart.draw()}.

    @ivar _chart: The chart that we are giving a view of.  This chart
       may be modified; after it is modified, you should call
       C{update}.
    @ivar _source: The list of tokens that the chart spans.

    @ivar _root: The root window.
    @ivar _chart_canvas: The canvas we're using to display the chart
        itself. 
    @ivar _tree_canvas: The canvas we're using to display the tree
        that each edge spans.  May be None, if we're not displaying
        trees. 
    @ivar _source_canvas: The canvas we're using to display the source
        text.  May be None, if we're not displaying the source text.
    @ivar _edgetags: A dictionary mapping from edges to the tags of
        the canvas elements (lines, etc) used to display that edge.
        The values of this dictionary have the form 
        C{(linetag, rhstag1, dottag, rhstag2, lhstag)}.
    @ivar _treetags: A list of all the tags that make up the tree;
        used to erase the tree (without erasing the loclines).
    @ivar _chart_height: The height of the chart canvas.
    @ivar _source_height: The height of the source canvas.
    @ivar _tree_height: The height of the tree

    @ivar _text_height: The height of a text string (in the normal
        font). 

    @ivar _edgelevels: A list of edges at each level of the chart (the
        top level is the 0th element).  This list is used to remember
        where edges should be drawn; and to make sure that no edges
        are overlapping on the chart view.

    @ivar _edgeselection: The currently selected edge; or None.
    
    @ivar _unitsize: Pixel size of one unit (from the location).  This
       is determined by the span of the chart's location, and the
       width of the chart display canvas.

    @ivar _fontsize: The current font size

    @ivar _marks: A dictionary from edges to marks.  Marks are
        strings, specifying colors (e.g. 'green').
    """
    
    _TOK_SPACING = 10
    _MARGIN = 10
    _TREE_LEVEL_SIZE = 12
    _CHART_LEVEL_SIZE = 40
    
    def __init__(self, chart, source=None, root=None, **kw):
        """
        Construct a new C{Chart} display.
        
        @type source: C{list} of C{Token}
        @param source: The list of Tokens that the chart spans.
        """
        # Process keyword args.
        draw_tree = kw.get('draw_tree', 0)
        draw_source = kw.get('draw_source', 1)
        self._fontsize = kw.get('fontsize', -12)
        
        if source is None:
            source = []
            draw_source = 0

        self._chart = chart
        self._source = source

        # Keep track of drawn edges
        self._edgelevels = []
        self._edgetags = {}

        # Keep track of which edges are marked.
        self._marks = {}

        # Keep track of a single "selected" edge
        self._edgeselection = None

        # Keep track of the tags used to draw the tree
        self._tree_tags = []

        # Put multiple edges on each level?
        self._compact = 1

        # If they didn't provide a main window, then set one up.
        if root is None:
            top = Tkinter.Tk()
            top.title('Chart View')
            def destroy1(e, top=top): top.destroy()
            def destroy2(top=top): top.destroy()
            top.bind('q', destroy1)
            b = Tkinter.Button(top, text='Done', command=destroy2)
            b.pack(side='bottom')
            self._root = top
        else:
            self._root = root

        # Create the chart canvas.
        (self._chart_sb, self._chart_canvas) = self._sb_canvas(self._root)
        self._chart_canvas['height'] = 400
        self._chart_canvas['closeenough'] = 15

        # Create the source canvas.
        if draw_source:
            cframe = Tkinter.Frame(self._root, relief='sunk', border=2)
            cframe.pack(fill='both', side='bottom')
            self._source_canvas = Tkinter.Canvas(cframe, height=50)
            self._source_canvas['background'] = '#e0e0e0'
            self._source_canvas.pack(fill='both')
            #self._source_canvas['height'] = self._source_height
        else:
            self._source_canvas = None

        # Create the tree canvas.
        if draw_tree:
            (sb, canvas) = self._sb_canvas(self._root, 'n', 'x')
            (self._tree_sb, self._tree_canvas) = (sb, canvas)
            self._tree_canvas['height'] = 200
        else:
            self._tree_canvas = None

        # Add an (optional?) view of the chart as a matrix.
        self._smallmatrix = SmallChartMatrixView(self._root, self._chart,
                                                 self._source)
        
        # Do some analysis to figure out how big the window should be
        self._analyze()
        self.draw()     # ***** why do I need this draw?
        self._resize()
        self._grow()

        # Set up the configure callback, which will be called whenever
        # the window is resized.
        self._chart_canvas.bind('<Configure>', self._configure)

    def _sb_canvas(self, root, expand='y', 
                   fill='both', side='bottom'):
        """
        Helper for __init__: construct a canvas with a scrollbar.
        """
        cframe =Tkinter.Frame(root, relief='sunk', border=2)
        cframe.pack(fill=fill, expand=expand, side=side)
        canvas = Tkinter.Canvas(cframe, background='#e0e0e0')
        
        # Give the canvas a scrollbar.
        sb = Tkinter.Scrollbar(cframe, orient='vertical')
        sb.pack(side='right', fill='y')
        canvas.pack(side='left', fill=fill, expand='yes')

        # Connect the scrollbars to the canvas.
        sb['command']= canvas.yview
        canvas['yscrollcommand'] = sb.set

        return (sb, canvas)

    def scroll_up(self, *e):
        self._chart_canvas.yview('scroll', -1, 'units')

    def scroll_down(self, *e):
        self._chart_canvas.yview('scroll', 1, 'units')

    def page_up(self, *e):
        self._chart_canvas.yview('scroll', -1, 'pages')

    def page_down(self, *e):
        self._chart_canvas.yview('scroll', 1, 'pages')

    def _grow(self):
        """
        Grow the window, if necessary
        """
        # Grow, if need-be
        loc = self._chart.loc()
        width = max(int(self._chart_canvas['width']),
                    (loc.end() - loc.start())
                    * self._unitsize +
                    ChartView._MARGIN * 2 )
        width = int(self._chart_canvas['width'])
        #self._chart_canvas['width'] = width # broken??
        unitwidth = loc.end() - loc.start()
        self._unitsize = (width - 2*ChartView._MARGIN) / unitwidth

        # Reset the height for the source window.
        if self._source_canvas is not None:
            self._source_canvas['height'] = self._source_height
        
    def set_font_size(self, size):
        self._fontsize = -abs(size)
        self._analyze()
        self._grow()
        self.draw()
            
    def get_font_size(self):
        return abs(self._fontsize)
            
    def _configure(self, e):
        """
        The configure callback.  This is called whenever the window is
        resized.  It is also called when the window is first mapped.
        It figures out the unit size, and redraws the contents of each
        canvas.
        """
        # This breaks things, but I still don't have this quite right:
        if abs(int(self._chart_canvas['width'])-e.width) > 5:
            self._chart_canvas['width'] = e.width
        loc = self._chart.loc()
        unitwidth = loc.end() - loc.start()
        self._unitsize = (e.width - 2*ChartView._MARGIN) / unitwidth
        self.draw()

    def update(self, chart=None, source=None):
        """
        Draw any edges that have not been drawn.  This is typically
        called when a after modifies the canvas that a CanvasView is
        displaying.  C{update} will cause any edges that have been
        added to the chart to be drawn.

        If update is given a C{chart} argument, then it will replace
        the current chart with the given chart.
        """
        if chart is not None:
            if source is not None: self._source = source
            self._chart = chart
            self._smallmatrix.update(self._chart, source)
            self._edgelevels = []
            self._edgeselection = None
            self._analyze()
            self._grow()
            self.draw()
            self._resize()
            # Update the matrix rep.
            self._smallmatrix.update(self._chart)
        else:
            edges = self._chart.edges()
            edges.sort()
            for edge in edges:
                if not self._edgetags.has_key(edge):
                    self._add_edge(edge)
            self._resize()


    def _edge_conflict(self, edge, lvl):
        """
        Return 1 if the given edge overlaps with any edge on the given
        level.  This is used by _add_edge to figure out what level a
        new edge should be added to.
        """
        loc = edge.loc()
        for otheredge in self._edgelevels[lvl]:
            otherloc = otheredge.loc()
            if loc.overlaps(otherloc):
                return 1
            if ((len(loc) == 0 or len(otherloc)==0) and
                loc.start() == otherloc.start()):
                return 1
        return 0

    def _analyze_edge(self, edge):
        """
        Given a new edge, recalculate:

            - _text_height
            - _unitsize (if the edge text is too big for the current
              _unitsize, then increase _unitsize)
        """
        c = self._chart_canvas

        if isinstance(edge, ProductionEdge):
            lhs = edge.prod().lhs()
            rhselts = []
            for elt in edge.prod().rhs():
                if isinstance(elt, Nonterminal):
                    rhselts.append(str(elt.symbol()))
                else:
                    rhselts.append(repr(elt))
            rhs = ' '.join(rhselts)
        else:
            lhs = edge.lhs()
            rhs = ''
            
        for s in (lhs, rhs):
            tag = c.create_text(0,0, text=s,
                                font=('helvetica', self._fontsize, 'bold'),
                                anchor='nw', justify='left')
            bbox = c.bbox(tag)
            c.delete(tag)
            width = bbox[2] #+ ChartView._TOK_SPACING
            edgelen = max(len(edge.loc()), 1)
            self._unitsize = max(self._unitsize, width/edgelen)
            self._text_height = max(self._text_height, bbox[3] - bbox[1])
    
    def _add_edge(self, edge, minlvl=0):
        """
        Add a single edge to the ChartView:

            - Call analyze_edge to recalculate display parameters
            - Find an available level
            - Call _draw_edge
        """
        if self._edgetags.has_key(edge): return
        self._analyze_edge(edge)
        self._grow()

        if not self._compact and isinstance(edge, ProductionEdge):
            self._edgelevels.append([edge])
            lvl = len(self._edgelevels)-1
            self._draw_edge(edge, lvl)
            self._chart_height = (lvl+2)*self._chart_level_size
            return

        # Figure out what level to draw the edge on.
        lvl = 0
        while 1:
            # If this level doesn't exit yet, create it.
            if lvl >= len(self._edgelevels):
                self._edgelevels.append([])
                self._chart_height = (lvl+2)*self._chart_level_size

            # Check if we can fit the edge in this level.
            if lvl>=minlvl and not self._edge_conflict(edge, lvl):
                # Go ahead and draw it.
                self._edgelevels[lvl].append(edge)
                break

            # Try the next level.
            lvl += 1

        self._draw_edge(edge, lvl)

    def view_edge(self, edge):
        level = None
        for i in range(len(self._edgelevels)):
            if edge in self._edgelevels[i]:
                level = i
                break
        if level == None: return
        
        # Try to view the new edge..
        y = (level+1) * self._chart_level_size
        dy = self._text_height + 10
        self._chart_canvas.yview('moveto', 1.0)
        if self._chart_height != 0:
            self._chart_canvas.yview('moveto',
                                     float(y-dy)/self._chart_height)

    def _draw_edge(self, edge, lvl):
        """
        Draw a single edge on the ChartView.
        """
        c = self._chart_canvas
        loc = edge.loc()
        
        # Draw the arrow.
        x1 = (loc.start() * self._unitsize + ChartView._MARGIN)
        x2 = (loc.end() * self._unitsize + ChartView._MARGIN)
        if x2 == x1: x2 += max(4, self._unitsize/5)
        y = (lvl+1) * self._chart_level_size
        linetag = c.create_line(x1, y, x2, y, arrow='last', width=3)

        # Draw a label for the edge.
        if isinstance(edge, ProductionEdge):
            rhs = []
            for elt in edge.prod().rhs():
                if isinstance(elt, Nonterminal):
                    rhs.append(str(elt.symbol()))
                else:
                    rhs.append(repr(elt))
            pos = edge.dotpos()
        else:
            rhs = []
            pos = 0
            
        rhs1 = ' '.join(rhs[:pos])
        rhs2 = ' '.join(rhs[pos:])
        rhstag1 = c.create_text(x1+3, y, text=rhs1,
                                font=('helvetica', self._fontsize),
                                anchor='nw')
        dotx = c.bbox(rhstag1)[2] + 6
        doty = (c.bbox(rhstag1)[1]+c.bbox(rhstag1)[3])/2
        dottag = c.create_oval(dotx-2, doty-2, dotx+2, doty+2)
        rhstag2 = c.create_text(dotx+6, y, text=rhs2,
                                font=('helvetica', self._fontsize),
                                anchor='nw')
        lhstag =  c.create_text((x1+x2)/2, y, text=str(edge.lhs()),
                                anchor='s',
                                font=('helvetica', self._fontsize, 'bold'))

        # Keep track of the edge's tags.
        self._edgetags[edge] = (linetag, rhstag1,
                                dottag, rhstag2, lhstag)

        # Register a callback for clicking on the edge.
        def cb(event, self=self, edge=edge): self.select_edge(edge)
        c.tag_bind(rhstag1, '<Button-1>', cb)
        c.tag_bind(rhstag2, '<Button-1>', cb)
        c.tag_bind(linetag, '<Button-1>', cb)
        c.tag_bind(dottag, '<Button-1>', cb)
        c.tag_bind(lhstag, '<Button-1>', cb)

        self._color_edge(edge)
        
    def _color_edge(self, edge, linecolor=None, textcolor=None):
        """
        Color in an edge with the given colors.
        If no colors are specified, use intelligent defaults
        (dependant on selection, etc.)
        """
        if not self._edgetags.has_key(edge): return
        c = self._chart_canvas

        if linecolor is not None and textcolor is not None:
            if self._marks.has_key(edge):
                linecolor = self._marks[edge]
            tags = self._edgetags[edge]
            c.itemconfig(tags[0], fill=linecolor)
            c.itemconfig(tags[1], fill=textcolor)
            c.itemconfig(tags[2], fill=textcolor,
                         outline=textcolor)
            c.itemconfig(tags[3], fill=textcolor)
            c.itemconfig(tags[4], fill=textcolor)
            self._smallmatrix.color_edge(edge, linecolor)
            return
        else:
            if edge is self._edgeselection:
                self._color_edge(edge, '#f00', '#800')
            elif (edge.complete() and edge.loc() == self._chart.loc()):
                self._color_edge(edge, '#084', '#042')
            elif isinstance(edge, TokenEdge):
                self._color_edge(edge, '#48c', '#246')
            else:
                self._color_edge(edge, '#00f', '#008')

    def mark_edge(self, edge, mark='#0df'):
        """
        Mark an edge
        """
        self._marks[edge] = mark
        self._color_edge(edge)

    def unmark(self, edge=None):
        """
        Unmark an edge (or all edges)
        """
        if edge == None:
            old_marked_edges = self._marks.keys()
            self._marks = {}
            for edge in old_marked_edges:
                self._color_edge(edge)
        else:
            del self._marks[edge]
            self._color_edge(edge)
                
    def select_edge(self, edge):
        """
        If the given edge is not currently selected, then unselect the
        old selected edge, and select the given edge.  If edge is
        currently selected, then deselect it.  If edge is None, then
        unselect the old edge.
        """
        c = self._chart_canvas

        # Unselect the old edge.
        oldedge = self._edgeselection
        if oldedge is not None:
            # If you select an already selected edge, then deselect it.
            if (edge and oldedge == edge): edge = None
            
            if not self._edgetags.has_key(oldedge):
                print 'Warning: _edgeselection is inconsistant'
                self._edgeselection = None
                return
            self._edgeselection = None
            self._color_edge(oldedge)

        if edge is None:
            self._edgeselection = None
            # Erase any old tree.
            if self._tree_canvas: self._draw_tree()
            return

        # Select the new edge.
        self._edgeselection = edge
        self._color_edge(edge)

        # Draw the tree for the newly selected edge.
        if self._tree_canvas: self._draw_tree()
            
    def _analyze(self):
        """
        Analyze the source string, to figure out how big a unit needs 
        to be, How big the tree should be, etc.
        """
        # Figure out the text height and the unit size.
        unitsize = 70 # min unitsize
        text_height = 0
        c = self._chart_canvas

        # Check against all tokens
        for tok in self._source:
            tag = c.create_text(0,0, text=repr(tok.type()),
                                font=('helvetica', self._fontsize),
                                anchor='nw', justify='left')
            bbox = c.bbox(tag)
            c.delete(tag)
            width = bbox[2] + ChartView._TOK_SPACING
            unitsize = max(width/len(tok.loc()), unitsize)
            text_height = max(text_height, bbox[3] - bbox[1])

        self._unitsize = unitsize
        self._text_height = text_height
        self._source_height = (self._text_height +
                               2*ChartView._MARGIN)

        # Check against edges.
        for edge in self._chart.edges():
            self._analyze_edge(edge)

        # Size of chart levels
        self._chart_level_size = self._text_height * 2.5
        
        # Default tree size..
        self._tree_height = (3 * (ChartView._TREE_LEVEL_SIZE +
                                  self._text_height))
        self._chart_height = 0

        # Resize the scrollregions.
        self._resize()

    def _resize(self):
        """
        Update the scroll-regions for each canvas.  This ensures that
        everything is within a scroll-region, so the user can use the
        scrollbars to view the entire display.  This does I{not}
        resize the window.
        """
        c = self._chart_canvas

        # Reset the chart scroll region
        width = ( (self._chart._loc.end() -
                   self._chart._loc.start()) * self._unitsize +
                  ChartView._MARGIN * 2 )
        height = self._chart_height
        c['scrollregion']=(0,0,width,height)

        # Reset the tree scroll region
        if self._tree_canvas:
            self._tree_canvas['scrollregion'] = (0, 0, width,
                                                 self._tree_height)
            
        #if int(self._tree_canvas['height']) > self._tree_height:
        #    self._tree_canvas['height'] = self._tree_height

        # Reset the height for the source window.
        #self._source_canvas['height'] = self._source_height
                                         
    def _draw_loclines(self):
        """
        Draw location lines.  These are vertical gridlines used to
        show where each location unit is.
        """
        c1 = self._tree_canvas
        c2 = self._source_canvas
        c3 = self._chart_canvas
        margin = ChartView._MARGIN
        self._loclines = []
        for i in range(self._chart._loc.start()-1,
                       self._chart._loc.end()+1):
            x = i*self._unitsize + margin

            if c1:
                t1=c1.create_line(x, 0, x, 5000)
                c1.tag_lower(t1)
            if c2:
                t2=c2.create_line(x, 0, x, self._source_height)
                c2.tag_lower(t2)
            t3=c3.create_line(x, 0, x, 5000)
            c3.tag_lower(t3)
            t4=c3.create_text(x+2, 0, text=`i`, anchor='nw',
                              font=('helvetica', self._fontsize))
            c3.tag_lower(t4)
            #if i % 4 == 0:
            #    if c1: c1.itemconfig(t1, width=2, fill='gray60')
            #    if c2: c2.itemconfig(t2, width=2, fill='gray60')
            #    c3.itemconfig(t3, width=2, fill='gray60')
            if i % 2 == 0:
                if c1: c1.itemconfig(t1, fill='gray60')
                if c2: c2.itemconfig(t2, fill='gray60')
                c3.itemconfig(t3, fill='gray60')
            else:
                if c1: c1.itemconfig(t1, fill='gray80')
                if c2: c2.itemconfig(t2, fill='gray80')
                c3.itemconfig(t3, fill='gray80')

    def selected_edge(self):
        """Return the currently selected edge."""
        return self._edgeselection

    def _draw_source(self):
        """Draw the source string."""
        if len(self._source) == 0: return
        c = self._source_canvas
        margin = ChartView._MARGIN
        y = ChartView._MARGIN
        
        for tok in self._source:
            x1 = tok.loc().start() * self._unitsize + margin
            x2 = tok.loc().end() * self._unitsize + margin
            x = (x1+x2)/2
            tag = c.create_text(x, y, text=repr(tok.type()),
                                font=('helvetica', self._fontsize),
                                anchor='n', justify='left')
            bbox = c.bbox(tag)
            rt=c.create_rectangle(x1+2, bbox[1]-(ChartView._TOK_SPACING/2),
                                  x2-2, bbox[3]+(ChartView._TOK_SPACING/2),
                                  fill='#f0f0f0', outline='#f0f0f0')
            c.tag_lower(rt)

    def _draw_tree(self):
        """Draw the syntax tree for the selected edge"""
        for tag in self._tree_tags:
            self._tree_canvas.delete(tag)
        edge = self._edgeselection
        if not edge: return
        if edge.structure() is None: return
        if not isinstance(edge.structure(), TreeToken): return

        treetok = edge.structure()
        
        if (isinstance(edge, ProductionEdge) and
            not edge.complete()):
            rhs = edge.prod().rhs()
            dotpos = edge.dotpos()
            
            self._draw_treetok(treetok, edge.loc(), 0, len(rhs)-dotpos)
            
            # Draw dotted lines to unexpanded children.
            c = self._tree_canvas
            x1 = edge.loc().start() * self._unitsize + ChartView._MARGIN
            x2 = ((edge.loc().end()+len(rhs)-dotpos) * self._unitsize +
                  ChartView._MARGIN)
            x = (x1+x2)/2
            childx = ((edge.loc().end()+0.5) * self._unitsize +
                      ChartView._MARGIN)
            
            childy = ChartView._TREE_LEVEL_SIZE + self._text_height
            for node in rhs[dotpos:]:
                tag = c.create_line(x, self._text_height, childx, childy,
                              width=2, fill='#048', dash='2 3')
                self._tree_tags.append(tag)
                # Draw node label.
                tag = c.create_text(childx, childy, anchor='n', fill='#042',
                                    justify='center', text=str(node), 
                                    font=('helvetica', self._fontsize, 'bold'))
                self._tree_tags.append(tag)
                childx += self._unitsize
                
            y = ChartView._TREE_LEVEL_SIZE + self._text_height
        else:
            self._draw_treetok(treetok, edge.loc(), 0)
                      
        self._tree_height = ((treetok.height()-1) *
                             (ChartView._TREE_LEVEL_SIZE +
                              self._text_height))
        self._resize()

    def _draw_treetok(self, treetok, loc, depth, offset=0):
        """
        Draw a treetoken.  Return the center x of this treetok.
        Return a negative x if treetok is not expanded.
        """
        if loc is None:
            print 'Warning: %r has location None' % treetok
        c = self._tree_canvas
        margin = ChartView._MARGIN
        x1 = loc.start() * self._unitsize + margin
        x2 = (loc.end()+offset) * self._unitsize + margin
        x = (x1+x2)/2
        y = depth * (ChartView._TREE_LEVEL_SIZE + self._text_height)
        tag = c.create_text(x, y, anchor='n', justify='center',
                            text=str(treetok.node()), fill='#042',
                            font=('helvetica', self._fontsize, 'bold'))
        self._tree_tags.append(tag)

        for child in treetok:
            if isinstance(child, TreeToken):
                childx = self._draw_treetok(child, child.loc(), depth+1)
                tag = c.create_line(x, y + self._text_height, childx,
                                    y + ChartView._TREE_LEVEL_SIZE +
                                    self._text_height, width=2,
                                    fill='#084')
            else:
                tag = c.create_line(x, y + self._text_height, x, 5000,
                                    width=2, fill='#084')
            self._tree_tags.append(tag)
        return x
            
    def draw(self):
        """
        Draw everything (from scratch).
        """
        if self._tree_canvas:
            self._tree_canvas.delete('all')
            self._draw_tree()

        if self._source_canvas:
            self._source_canvas.delete('all')
            self._draw_source()

        self._chart_canvas.delete('all')
        self._edgetags = {}
        
        # Redraw any edges we erased.
        for lvl in range(len(self._edgelevels)):
            for edge in self._edgelevels[lvl]:
                self._draw_edge(edge, lvl)

        # Add any new edges
        edges = self._chart.edges()
        edges.sort()
        for edge in edges:
            self._add_edge(edge)

        self._draw_loclines()

#//////////////////////////////////////////////////////////////////////
# Edge Rules
#//////////////////////////////////////////////////////////////////////
# These version of the chart rules only apply to a specific edge.
# This lets the user select an edge, and then apply a rule.

class TopDownEdgeRule(ChartRuleI):
    def __init__(self, edge): self._edge = edge
    def apply(self, chart, grammar):
        if self._edge.complete(): return []
        edges = []
        for production in grammar.productions():
            if production.lhs() == self._edge.next():
                loc = self._edge.loc().end_loc()
                edges.append(self_loop_edge(production, loc))
        return edges
    def __str__(self): return 'Top-down Rule'

class FundamentalEdgeRule(ChartRuleI):
    def __init__(self, edge): self._edge = edge
    def apply(self, chart, grammar):
        edges = []
        if self._edge.complete():
            for edge2 in chart.incomplete_edges():
                if (edge2.next() == self._edge.lhs() and
                    edge2.end() == self._edge.start()):
                    edges.append(fr_edge(edge2, self._edge))
        else:
            for edge2 in chart.complete_edges():
                if (self._edge.next() == edge2.lhs() and
                    self._edge.end() == edge2.start()):
                    edges.append(fr_edge(self._edge, edge2))
        return edges
    def __str__(self): return 'Fundamental Rule'

class BottomUpEdgeRule(ChartRuleI):
    def __init__(self, edge): self._edge = edge
    def apply(self, chart, grammar):
        edges = []
        for production in grammar.productions():
            if self._edge.lhs() == production.rhs()[0]:
                loc = self._edge.loc().start_loc()
                edges.append(self_loop_edge(production, loc))
        return edges
    def __str__(self): return 'Bottom Up Rule'

#//////////////////////////////////////////////////////////////////////
# Chart Demo
#//////////////////////////////////////////////////////////////////////

class ChartDemo:
    RULENAME = {'FundamentalRule': 'Fundamental Rule',
                'FundamentalEdgeRule': 'Fundamental Rule',

                'TopDownInitRule': 'Top-down Initialization',
                
                'TopDownRule': 'Top-down Rule',
                'TopDownEdgeRule': 'Top-down Rule',
                
                'BottomUpRule': 'Bottom-up Rule',
                'BottomUpEdgeRule': 'Bottom-up Rule',
                }
    
    def __init__(self, grammar, text, title='Chart Parsing Demo'):

        self._root = None
        try:
            self._lexiconview = self._grammarview = None
            # Create the root window.
            self._root = Tkinter.Tk()
            self._root.title(title)
            self._root.bind('q', self.destroy)
            buttons2 = Tkinter.Frame(self._root)
            buttons2.pack(side='bottom', fill='none')
            buttons1 = Tkinter.Frame(self._root)
            buttons1.pack(side='bottom', fill='x')

    
            self._grammar = grammar
            self._tok_sent = text
            self._cp = SteppingChartParser(self._grammar)
            self._cp.initialize(self._tok_sent)

            # Keep track of whether we're stepping & animating.
            self._step = Tkinter.IntVar(self._root)
            self._step.set(1)
            self._animate = Tkinter.IntVar(self._root)
            self._animate.set(2)

            self._chart = self._cp.chart()
            self._cv = ChartView(self._chart, self._tok_sent,
                                 self._root, draw_tree=1, draw_source=1)

            ruletxt = 'Last edge generated by:'
            self._rulelabel1 = Tkinter.Label(buttons1,text=ruletxt)
            self._rulelabel2 = Tkinter.Label(buttons1, width=30,
                                            font=('helvetica', 30, 'bold'),
                                            relief='groove', anchor='w')
            self._rulelabel1.pack(side='left')
            self._rulelabel2.pack(side='left')
            step = Tkinter.Checkbutton(buttons1, variable=self._step,
                                       text='Step')
            step.pack(side='right')

            ## Set up buttons for rules & strategies
            Tkinter.Button(buttons2, text='Top Down\nStrategy',
                           background='#90c0d0', foreground='black',
                           command=self.top_down_strategy).pack(side='left')
            Tkinter.Button(buttons2, text='Bottom Up\nStrategy',
                           background='#90c0d0', foreground='black',
                           command=self.bottom_up_strategy).pack(side='left')
            Tkinter.Button(buttons2, text='Top Down Init\nRule',
                           background='#90f090', foreground='black',
                           command=self.top_down_init).pack(side='left')
            Tkinter.Button(buttons2, text='Top down\nRule',
                           background='#90f090', foreground='black',
                           command=self.top_down).pack(side='left')
            Tkinter.Button(buttons2, text='Bottom Up Init\nRule',
                           background='#90f090', foreground='black',
                           command=self.bottom_up_init).pack(side='left')
            Tkinter.Button(buttons2, text='Fundamental\nRule',
                           background='#90f090', foreground='black',
                           command=self.fundamental).pack(side='left')
    
            # For animations..
            self._animating = 0  # are we animating right now?
        
            # Initialize the rule-label font.
            size = self._cv.get_font_size()
            self._rulelabel1.configure(font=('helvetica', -size, 'bold'))
            self._rulelabel2.configure(font=('helvetica', -size, 'bold'))

            # Set up a menu bar.
            self._init_menubar()

            # Set up keyboard bindings.
            self._init_bindings()

        except:
            print 'Error creating Tree View'
            self.destroy()
            raise

    def _init_bindings(self):
        self._root.bind('<Up>', self._cv.scroll_up)
        self._root.bind('<Down>', self._cv.scroll_down)
        self._root.bind('<Prior>', self._cv.page_up)
        self._root.bind('<Next>', self._cv.page_down)
        self._root.bind('<Control-q>', self.destroy)
        self._root.bind('<Control-x>', self.destroy)
        self._root.bind('<F1>', self.help)
        
        self._root.bind('<Control-s>', self.save)
        self._root.bind('<Control-o>', self.load)
        self._root.bind('<Control-r>', self.reset)
        
        self._root.bind('t', self.top_down_strategy)
        self._root.bind('b', self.bottom_up_strategy)
        self._root.bind('f', self.fundamental)
        self._root.bind('<Control-g>', self.edit_grammar)
        self._root.bind('<Control-t>', self.edit_sentence)

        # Animation speed control
        self._root.bind('-', lambda e,a=self._animate:a.set(1))
        self._root.bind('=', lambda e,a=self._animate:a.set(2))
        self._root.bind('+', lambda e,a=self._animate:a.set(3))

        # Step control
        self._root.bind('s', lambda e,s=self._step:s.set(not s.get()))

    def _init_menubar(self):
        menubar = Tkinter.Menu(self._root)
        
        filemenu = Tkinter.Menu(menubar, tearoff=0)
        filemenu.add_command(label='Save Chart', underline=0,
                             command=self.save, accelerator='Ctrl-s')
        filemenu.add_command(label='Open Chart', underline=0,
                             command=self.load, accelerator='Ctrl-o')
        filemenu.add_command(label='Reset Chart', underline=0,
                             command=self.reset, accelerator='Ctrl-r')
        filemenu.add_command(label='Exit', underline=1,
                             command=self.destroy,
                             accelerator='Ctrl-x')
        menubar.add_cascade(label='File', underline=0, menu=filemenu)
        
        editmenu = Tkinter.Menu(menubar, tearoff=0)
        editmenu.add_command(label='Edit Grammar', underline=5,
                             command=self.edit_grammar,
                             accelerator='Ctrl-g')
        editmenu.add_command(label='Edit Text', underline=5,
                             command=self.edit_sentence,
                             accelerator='Ctrl-t')
        menubar.add_cascade(label='Edit', underline=0, menu=editmenu)
        
        rulemenu = Tkinter.Menu(menubar, tearoff=0)
        rulemenu.add_command(label='Top Down Strategy', underline=0,
                             command=self.top_down_strategy,
                             accelerator='t')
        rulemenu.add_command(label='Bottom Up Strategy', underline=0,
                             command=self.bottom_up_strategy,
                             accelerator='b')
        rulemenu.add_separator()
        rulemenu.add_command(label='Bottom Up Init Rule',
                             command=self.bottom_up_init)
        rulemenu.add_command(label='Top Down Init Rule',
                             command=self.top_down_init)
        rulemenu.add_command(label='Top Down Rule',
                             command=self.top_down_init)
        rulemenu.add_command(label='Fundamental Rule', underline=0,
                             command=self.fundamental, accelerator='f')
        menubar.add_cascade(label='Apply', underline=0, menu=rulemenu)

        animatemenu = Tkinter.Menu(menubar, tearoff=0)
        animatemenu.add_checkbutton(label="Step", underline=0,
                                    variable=self._step,
                                    accelerator='s')
        animatemenu.add_separator()
        animatemenu.add_radiobutton(label="No Animation", underline=0,
                                    variable=self._animate, value=0)
        animatemenu.add_radiobutton(label="Slow Animation", underline=0,
                                    variable=self._animate, value=1,
                                    accelerator='-')
        animatemenu.add_radiobutton(label="Normal Animation", underline=0,
                                    variable=self._animate, value=2,
                                    accelerator='=')
        animatemenu.add_radiobutton(label="Fast Animation", underline=0,
                                    variable=self._animate, value=3,
                                    accelerator='+')
        menubar.add_cascade(label="Animate", underline=1, menu=animatemenu)
        
        zoommenu = Tkinter.Menu(menubar, tearoff=0)
        self._size = Tkinter.IntVar(self._root)
        self._size.set(self._cv.get_font_size())
        zoommenu.add_radiobutton(label='Tiny', variable=self._size,
                                 underline=0, value=10, command=self.resize)
        zoommenu.add_radiobutton(label='Small', variable=self._size,
                                 underline=0, value=12, command=self.resize)
        zoommenu.add_radiobutton(label='Medium', variable=self._size,
                                 underline=0, value=14, command=self.resize)
        zoommenu.add_radiobutton(label='Large', variable=self._size,
                                 underline=0, value=18, command=self.resize)
        zoommenu.add_radiobutton(label='Huge', variable=self._size,
                                 underline=0, value=24, command=self.resize)
        menubar.add_cascade(label='Zoom', underline=0, menu=zoommenu)

        helpmenu = Tkinter.Menu(menubar, tearoff=0)
        helpmenu.add_command(label='About', underline=0,
                             command=self.about)
        helpmenu.add_command(label='Instructions', underline=0,
                             command=self.help, accelerator='F1')
        menubar.add_cascade(label='Help', underline=0, menu=helpmenu)
        
        self._root.config(menu=menubar)

    def help(self, *e):
        self._animating = 0
        # The default font's not very legible; try using 'fixed' instead. 
        try:
            ShowText(self._root, 'Help: Chart Parser Demo',
                     (__doc__).strip(), width=75, font='fixed')
        except:
            ShowText(self._root, 'Help: Chart Parser Demo',
                     (__doc__).strip(), width=75)

    def about(self, *e):
        ABOUT = ("NLTK Chart Parser Demo\n"+
                 "Written by Edward Loper")
        try:
            from tkMessageBox import Message
            Message(message=ABOUT, title='About: Chart Parser Demo').show()
        except:
            ShowText(self._root, 'About: Chart Parser Demo', ABOUT)

    def load(self, *e):
        "Load a chart from a pickle file"
        import pickle
        from tkFileDialog import askopenfilename
        ftypes = [('Pickle file', '.pickle'),
                  ('All files', '*')]
        filename = askopenfilename(filetypes=ftypes,
                                   defaultextension='.pickle')
        if not filename: return
        chart = pickle.load(open(filename, 'r'))
        self._chart = chart
        self._cv.update(chart)
        self._cp.set_chart(chart)

    def save(self, *e):
        "Save a chart to a pickle file"
        import pickle
        from tkFileDialog import asksaveasfilename
        ftypes = [('Pickle file', '.pickle'),
                  ('All files', '*')]
        filename = asksaveasfilename(filetypes=ftypes,
                                     defaultextension='.pickle')
        if not filename: return
        pickle.dump(self._chart, open(filename, 'w'))
            
    def resize(self):
        self._animating = 0
        size = self._size.get()
        self._cv.set_font_size(size)
        self._rulelabel1['font'] = ('helvetica', -size, 'bold')
        self._rulelabel2['font'] = ('helvetica', -size, 'bold')

    def edit_grammar(self, *e):
        CFGEditor(self._root, self._grammar, self.set_grammar)

    def set_grammar(self, grammar):
        self._grammar = grammar
        self._cp.set_grammar(grammar)

    def edit_sentence(self, *e):
        sentence = ' '.join([tok.type() for tok in self._tok_sent])
        title = 'Edit Text'
        EntryDialog(self._root, sentence, self.set_sentence, title)

    def set_sentence(self, sentence):
        self._tok_sent = WSTokenizer().tokenize(sentence)
        self.reset()

    # not used:
    def view_grammar(self, *e):
        self._animating = 0
        self._grammarview = ProductionView(self._grammar.productions(),
                                           'Grammar')

    def reset(self, *e):
        self._animating = 0
        self._cp = SteppingChartParser(self._grammar)
        self._cp.initialize(self._tok_sent)
        self._chart = self._cp.chart()
        self._cv.update(self._chart, self._tok_sent)

    def display_rule(self, rule):
        if rule == None:
            self._rulelabel2['text'] = ''
        else:
            name = ChartDemo.RULENAME.get(rule.__class__.__name__,
                                          rule.__class__.__name__)
            self._rulelabel2['text'] = name
            size = self._cv.get_font_size()
            #self._rulelabel1['font'] = ('helvetica', -size, 'bold')
            #self._rulelabel2['font'] = ('helvetica', -size, 'bold')
            
    def apply_strategy(self, strategy, edge_strategy=None):
        if self._animating:
            self._animating = 0
            return
        self.display_rule(None)
        self._cv.unmark()
        if self._step.get():
            edge = self._cv.selected_edge()
            if (edge is not None) and (edge_strategy is not None):
                if self._apply_strategy([edge_strategy(edge)]) is None:
                    # Unselect it (select_edge toggles selection)
                    self._cv.select_edge(edge)
            else:
                self._apply_strategy(strategy)
        else:
            if self._animate.get():
                self._animating = 1
                self._animate_strategy(strategy)
            else:
                while self._cp.step(strategy=strategy):
                    self._cv.update()

    def _animate_strategy(self, strategy, speed=1):
        if self._animating == 0: return
        new_edge = self._cp.step(strategy=strategy)
        self._cv.unmark()
        if new_edge is not None:
            self.display_rule(self._cp.current_chartrule())
            self._cv.update()
            self._cv.view_edge(new_edge)
            if self._animate.get() == 0 or self._step.get() == 1:
                return
            if self._animate.get() == 1:
                self._cv.select_edge(new_edge)
                self._root.after(3000, self._animate_strategy, strategy)
            elif self._animate.get() == 2:
                self._cv.select_edge(new_edge)
                self._root.after(1000, self._animate_strategy, strategy)
            else:
                self._cv.mark_edge(new_edge)
                self._root.after(20, self._animate_strategy, strategy)
        
    def _apply_strategy(self, strategy):
        new_edge = self._cp.step(strategy=strategy)
                               
        if new_edge is not None:
            self.display_rule(self._cp.current_chartrule())
            self._cv.update()
            self._cv.mark_edge(new_edge)
            self._cv.view_edge(new_edge)
        return new_edge

    _TOP_DOWN_INIT = [TopDownInitRule()]
    def top_down_init(self, *e):
        self.apply_strategy(ChartDemo._TOP_DOWN_INIT, None)

    _TOP_DOWN_S = [TopDownRule(), FundamentalRule(), TopDownInitRule()]
    def top_down_strategy(self, *e):
        self.apply_strategy(ChartDemo._TOP_DOWN_S, TopDownEdgeRule)

    _TOP_DOWN = [TopDownRule()]
    def top_down(self, *e):
        self.apply_strategy(ChartDemo._TOP_DOWN, TopDownEdgeRule)

    _FUNDAMENTAL = [FundamentalRule()]
    def fundamental(self, *e):
        self.apply_strategy(ChartDemo._FUNDAMENTAL, FundamentalEdgeRule)

    _BOTTOM_UP_INIT = [BottomUpRule()]
    def bottom_up_init(self, *e):
        self.apply_strategy(ChartDemo._BOTTOM_UP_INIT, BottomUpEdgeRule)

    _BOTTOM_UP_S = [FundamentalRule(), BottomUpRule()]
    def bottom_up_strategy(self, *e):
        self.apply_strategy(ChartDemo._BOTTOM_UP_S, BottomUpEdgeRule)
        
    def destroy(self, *args):
        if self._lexiconview: self._lexiconview.destroy()
        if self._grammarview: self._grammarview.destroy()
        if self._root is None: return
        self._root.destroy()
        self._root = None

    def mainloop(self, *args, **kwargs):
        """
        Enter the Tkinter mainloop.  This function must be called if
        this demo is created from a non-interactive program (e.g.
        from a secript); otherwise, the demo will close as soon as
        the script completes.
        """
        if in_idle(): return
        self._root.mainloop(*args, **kwargs)
        
def demo():
    nonterminals = 'S VP NP PP P N Name V Det'
    (S, VP, NP, PP, P, N, Name, V, Det) = [Nonterminal(s)
                                           for s in nonterminals.split()]
    
    grammar_rules1 = [
        CFGProduction(NP, Det, N),  CFGProduction(NP, NP, PP),
        CFGProduction(NP, 'John'),  CFGProduction(NP, 'I'), 
        CFGProduction(Det, 'the'),  CFGProduction(Det, 'my'),
        CFGProduction(N, 'dog'),    CFGProduction(Det, 'a'),
        CFGProduction(N, 'cookie'), CFGProduction(N, 'table'),
        CFGProduction(VP, VP, PP),  CFGProduction(VP, V, NP),
        CFGProduction(V, 'ate'),    CFGProduction(V, 'saw'),
        CFGProduction(S, NP, VP),   CFGProduction(PP, P, NP),
        CFGProduction(P, 'on'),     CFGProduction(P, 'under'),
        CFGProduction(VP, V),
        ]

    grammar = CFG(S, grammar_rules1)
    sent = 'John ate the cookie on the table'
    tok_sent = WSTokenizer().tokenize(sent)

    print 'grammar= ('
    for rule in grammar.productions():
        print '    ', repr(rule)+','
    print ')'
    print 'sentence = %r' % sent
    print 'Calling "ChartDemo(grammar, tok_sent)"...'
    ChartDemo(grammar, tok_sent).mainloop()

def demo2():
    nonterminals = 'NP Det N'
    (NP, Det, N) = [Nonterminal(s) for s in nonterminals.split()]
    toks = [Token('the', 0), Token('park', 1), Token('I', 2), Token('saw', 3)]
    tree = TreeToken('NP',
                     TreeToken('Det', toks[0]),
                     TreeToken('N', toks[1]))

    edge0 = TokenEdge(toks[0])
    edge1 = TokenEdge(toks[1])

    edge2 = ProductionEdge(CFGProduction(Det, 'the'),
                           tree[0], tree[0].loc(), 1)
    edge3 = ProductionEdge(CFGProduction(N, 'park'),
                           tree[1], tree[1].loc(), 1)

    edge4 = ProductionEdge(CFGProduction(NP, Det, N),
                           tree, tree.loc(), 2)
    
    chart = Chart(Location(0,4))#tree.loc())
    chart.insert(edge0)
    chart.insert(edge1)
    chart.insert(edge2)
    chart.insert(edge3)
    chart.insert(edge4)
    chart.draw(toks, draw_tree=1)

if __name__ == '__main__':
    demo()

